#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import os
import pipes
import platform
import select
import subprocess
import sys
import traceback

is_windows = True if platform.system().lower() == 'windows' else False

# Useful for very coarse version differentiation.
PY3 = sys.version_info[0] == 3

ERROR_LIST = frozenset((None, 'surrogate_or_strict'))

if PY3:
    import io

    StringIO = io.StringIO
    BytesIO = io.BytesIO

    def b(s):
        return s.encode("latin-1")
    securecat_text_type = str
    securecat_binary_type = bytes
    
else:
    def b(s):
        return s
    import StringIO

    BytesIO= StringIO.StringIO
    StringIO = StringIO.StringIO
    securecat_text_type = unicode
    securecat_binary_type = str
    

def to_native(change_obj, encoding='utf-8', errors=None, string_type='simplerepr'):
    if isinstance(change_obj, str):
        return change_obj

    try:
        codecs.lookup_error('surrogateescape')
        error_surrogateescape = True
    except LookupError:
        error_surrogateescape = False

    original_errors = errors
    if errors in ERROR_LIST:
        if error_surrogateescape:
            errors = 'surrogateescape'
        elif errors == 'surrogate_or_strict':
            errors = 'strict'
        else:
            errors = 'replace'

    if PY3:
        if isinstance(change_obj, bytes):
            return change_obj.decode(encoding, errors)
    else:
        if isinstance(change_obj, unicode):
            try:
                return change_obj.encode(encoding, errors)
            except UnicodeEncodeError:
                if original_errors in (None, 'surrogate_then_replace'):
                    # Slow but works
                    return_string = change_obj.encode('utf-8', 'surrogateescape').decode('utf-8', 'replace')
                    return return_string.encode(encoding, 'replace')
                raise

    if string_type == 'simplerepr':
        try:
            change_value = str(change_obj)
        except UnicodeError:
            try:
                change_value = repr(change_obj)
            except UnicodeError:
                return ''
    elif string_type == 'passthru':
        return change_obj
    elif string_type == 'empty':
        return ''
    else :
        raise TypeError('Invalid value %s,value must be a string type' % string_type)

    return to_native(change_value, encoding, errors)


def read_pipes(ppipes, rfds, file_descriptor):
    str_data = ''
    if file_descriptor in rfds:
        str_data = os.read(file_descriptor.fileno(), 9000)
        if isinstance(str_data, (securecat_binary_type, securecat_text_type)):
            str_data = to_native(str_data)
        if str_data == '':
            ppipes.remove(file_descriptor)

    return str_data


def run_command(commands, check_rc=True, close_fds=True, executable=None, encoding='utf-8',
                errors='surrogate_or_strict'):
    shell = True
    ret_code = 0
    msg = None

    if not isinstance(commands, (list, securecat_binary_type, securecat_text_type)):
        msg = "run commands must be list or string"
        return (-1, 0, msg, '')
    elif isinstance(commands, list):
        commands = " ".join([pipes.quote(x) for x in commands])

    try:
        cmd = subprocess.Popen(commands, executable=executable, shell=shell, close_fds=close_fds, stdin=None,
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout = ''
        stderr = ''
        rpipes = [cmd.stdout, cmd.stderr]

        if not is_windows:
            while True:
                rfds, wfds, efds = select.select(rpipes, [], rpipes, 1)
                del wfds
                del efds
                stdout += read_pipes(rpipes, rfds, cmd.stdout)
                stderr += read_pipes(rpipes, rfds, cmd.stderr)
                # break out if no pipes are left to read or read completely
                if (not rpipes or not rfds) and cmd.poll() is not None:
                    break
                elif not rpipes and cmd.poll() is None:
                    cmd.wait()
                    break
        else:
            # shell module execute command on windows
            stdout = cmd.stdout.read().strip()
            stderr = cmd.stderr.read().strip()
            stdout = to_native(stdout)
            stderr = to_native(stderr)

        cmd.stdout.close()
        cmd.stderr.close()

        ret_code = cmd.returncode
    except (OSError, IOError):
        e = sys.exc_info()[1]
        return (-1, e.errno, to_native(e), traceback.format_exc())
    except Exception:
        e = sys.exc_info()[1]
        return (-1, 257, to_native(e), traceback.format_exc())

    if ret_code != 0 and check_rc:
        return (-1, ret_code, stdout, stderr)

    if encoding is not None:
        return (0, ret_code, to_native(stdout, encoding=encoding, errors=errors),
                to_native(stderr, encoding=encoding, errors=errors))
    return (0, ret_code, stdout, stderr)

